from avocado.query.operators import ContainerTypeOperator, registry

class RequireAll(ContainerTypeOperator):
    lookup = 'all'
    join_operator = 'and'
    short_name = 'require all'
    verbose_name = 'requires all of'


class NotAll(RequireAll):
    short_name = 'not all'
    verbose_name = 'can not have'
    negated = True


class Only(RequireAll):
    short_name = 'only be'
    verbose_name = 'can only be'
    lookup = 'only'


registry.register(RequireAll, RequireAll.uid)
registry.register(NotAll, NotAll.uid)
registry.register(Only, Only.uid)



from django.conf.urls.defaults import url, patterns, include
from .settings import VOCAB_FIELDS

override_fields = '|'.join([str(f) for f in VOCAB_FIELDS])

urlpatterns = patterns('',
    url(r'', include(patterns('',
        url(r'^fields/(?P<pk>{0})/values/$'.format(override_fields),
            'vocab.resources.ItemsResource', name='items'),
       url(r'^fields/(?P<pk>{0})/values/(?P<item_pk>\d+)/$'.format(override_fields),
           'vocab.resources.ItemResource', name='item'),
       url(r'^fields/(?P<pk>{0})/values/(?P<item_pk>\d+)/values/$'.format(override_fields),
           'vocab.resources.ItemsResource', name='items'),
    ), namespace='vocab'))
)



from django.db import connection
from django.core.exceptions import ImproperlyConfigured
from avocado.query.translators import Translator
from modeltree.tree import trees

qn = connection.ops.quote_name

class VocabularyTranslator(Translator):
    through_model = None

    operators = ('in', '-in', '-all', 'all', 'only')

    def __init__(self, *args, **kwargs):
        if not self.through_model:
            raise ImproperlyConfigured('Translator requires `through_model` attribute be defined')
        super(VocabularyTranslator, self).__init__(*args, **kwargs)

    def translate(self, field, roperator, rvalue, tree, **context):
        # Validate the operator and value
        operator, value = self.validate(field, roperator, rvalue, tree, **context)

        # Build the default language representation
        language = self.language(field, operator, value, tree=tree, **context)

        through = self.through_model
        if operator.uid == 'in':
            subquery = through.objects.requires_any(value)
        elif operator.uid == '-in':
            subquery = through.objects.excludes_any(value)
        elif operator.uid == 'all':
            subquery = through.objects.requires_all(value)
        elif operator.uid == '-all':
            subquery = through.objects.excludes_all(value)
        elif operator.uid == 'only':
            subquery = through.objects.only(value)
        else:
            raise ImproperlyConfigured('Unknown operator for vocab translator')

        # Get all joins from the target tree model to the 'object' field on
        # the through model.
        tree = trees[tree]
        joins = tree.get_joins(through.objects.object_field.rel.to)

        # Collect where/tables information required for the `QuerySet.extra()`
        # method. See: https://docs.djangoproject.com/en/1.5/ref/models/querysets/#extra
        tables = []
        where = []

        for join in joins[1:]:
            conn = join.get('connection')

            tables.append(conn[1])

            kwargs = {
                'left': qn(conn[0]),
                'right': qn(conn[1]),
                'left_id': qn(conn[2]),
                'right_id': qn(conn[3]),
            }

            where.append('{left}.{left_id} = {right}.{right_id}'.format(**kwargs))

        where.append('{object_table}.{object_id} IN {subquery}'.format(**{
            'object_table': qn(through.objects.object_field.rel.to._meta.db_table),
            'object_id': qn(through.objects.object_field.rel.to._meta.pk.column),
            'subquery': subquery,
        }))

        # Extract description of each 'value' which will be the item object
        # itself.
        new_value = [unicode(v) for v in value]

        return {
            'id': field.pk,
            'operator': roperator,
            'value' : new_value,
            'cleaned_data':{
                'value': value,
                'operator': operator,
                'language': language,
            },
            'query_modifiers': {
                'condition': None,
                'annotations': None,
                'extra': {
                    'where': where,
                    'tables': tables,
                },
            }
        }



import functools
from django.db.models import Q
from django.core.urlresolvers import reverse
from avocado.models import DataField
from avocado.events import usage
from restlib2.resources import Resource
from serrano.resources.field.values import FieldValues
from preserialize.serialize import serialize


def item_posthook(instance, data, request, pk):
    "Post-serialize hook for item instances."
    uri = request.build_absolute_uri

    data.update({
        'label': unicode(instance),
        'value': instance.pk,
    })

    parent_kwargs = {'pk': pk}
    if instance.parent_id:
        parent_kwargs['item_pk'] = instance.parent_id

    # Add link to self and parent
    data['_links'] = {
        'self': {
            'href': uri(reverse('vocab:item', kwargs={
                'pk': pk,
                'item_pk': data['id'],
            }))
        },
        'parent': {
            'href': uri(reverse('vocab:items', kwargs=parent_kwargs))
        }
    }

    # If this item is not a terminal, add link to child nodes
    if not instance.terminal:
        data['_links']['children'] = {
            'href': uri(reverse('vocab:items', kwargs={
                'pk': pk,
                'item_pk': data['id'],
            }))
        }

    return data


class ItemBaseResource(Resource):
    def is_not_found(self, request, response, pk, item_pk=None):
        # Non-integer value
        try:
            pk = int(pk)
        except (ValueError, TypeError):
            return True

        # Field does not exist
        try:
            field = DataField.objects.get(pk=pk)
        except DataField.DoesNotExist:
            return True

        # If an item is specified, ensure it exists
        if item_pk:
            try:
                request.item = field.model.objects.get(pk=item_pk)
            except field.model.DoesNotExist:
                return True

        # Attach to request for downstream use
        request.instance = field

    def prepare(self, request, objects, pk, template=None):
        if template is None:
            template = {'fields': [':pk']}
        posthook = functools.partial(item_posthook, request=request, pk=pk)
        return serialize(objects, posthook=posthook, **template)


# TODO if/when serrano becomes more reusable, update and remove
# boilerplate code
class ItemsResource(ItemBaseResource, FieldValues):
    """Resource for vocab items. If no item is specified, items without a
    parent are returned, otherwise the children of the specified item are
    returned.

    This is a modified form of serrano.resources.FieldValues.
    """

    template = None

    def get_base_values(self, request, instance, params, item_pk=None):
        queryset = super(ItemsResource, self).get_base_values(request,
                    instance, params)
        queryset = queryset.filter(parent__pk=item_pk)
        return queryset

    def get_all_values(self, request, instance, params, item_pk=None):
        queryset = self.get_base_values(request, instance, params, item_pk)
        return self.prepare(request, queryset, instance.pk)

    def get_search_values(self, request, instance, params, item_pk=None):
        queryset = self.get_base_values(request, instance, params, item_pk)
        condition = Q()
        for field in instance.model.search_fields:
            condition = condition | Q(**{'{0}__icontains'.format(field): params['query']})
        queryset = queryset.filter(condition)
        return self.prepare(request, queryset, instance.pk)

    def get_random_values(self, request, instance, params, item_pk=None):
        queryset = self.get_base_values(request, instance, params, item_pk)
        queryset = queryset.order_by('?')[:params['random']]
        return self.prepare(request, queryset, instance.pk)

    def get(self, request, pk, item_pk=None):
        instance = request.instance
        params = self.get_params(request)

        if params['random']:
            return self.get_random_values(request, instance, params, item_pk)

        page = params['page']
        limit = params['limit']

        # If a query term is supplied, perform the icontains search
        if params['query']:
            usage.log('values', instance=instance, request=request, data={
                'query': params['query'],
                'item_pk': item_pk,
            })
            values = self.get_search_values(request, instance, params, item_pk)
        else:
            values = self.get_all_values(request, instance, params, item_pk)

        # No page specified, return everything
        if page is None:
            return values

        paginator = self.get_paginator(values, limit=limit)
        page = paginator.page(page)

        kwargs = {'pk': pk}
        if item_pk:
            kwargs['item_pk'] = item_pk
        path = reverse('vocab:items', kwargs=kwargs)
        links = self.get_page_links(request, path, page, extra=params)
        links['parent'] = {
            'href': request.build_absolute_uri(reverse('serrano:field',
                    kwargs={'pk': pk}))
        }

        return {
            'values': page.object_list,
            'limit': paginator.per_page,
            'num_pages': paginator.num_pages,
            'page_num': page.number,
            '_links': links,
        }


class ItemResource(ItemBaseResource):
    def get(self, request, pk, item_pk):
        return self.prepare(request, request.item, pk)



import re
from django.db import models, transaction, connection
from django.db.models.sql import RawQuery

qn = connection.ops.quote_name

# Prefix for the summed case statement values
COLUMN_PREFIX = 'sc'

# Case statement for determining if the lookup matches the item or the
# parent. Every record is evaluated
SUMMED_CASE_TEMPLATE = '''\
SUM(
    CASE
        WHEN {index_table}.{item_column} = {item_pk} THEN 1
        WHEN {index_table}.{parent_column} = {item_pk} THEN 1
        ELSE 0
    END
) AS {column_alias}
'''

# Sub-query which selects all object ids matching the conditions derived
# from the CASE statements.
PIVOT_QUERY_TEMPLATE = '''\
SELECT {object_column} FROM (
    SELECT {object_column}, {case_statements}
    FROM {through_table}
        INNER JOIN {index_table} ON ({through_table}.{item_column} = {index_table}.{item_id})
    GROUP BY {object_column}
) AS T WHERE {negate} ({where_condition})
'''


class ItemManager(models.Manager):
    def ancestors(self, pk, *args, **kwargs):
        "Returns a ``QuerySet`` containing all ancestors of this item."
        return self.get_query_set().get(pk=pk).ancestors(*args, **kwargs)

    def descendants(self, pk, *args, **kwargs):
        "Returns a ``QuerySet`` containing all descendants of this item."
        return self.get_query_set().get(pk=pk).descendants(*args, **kwargs)


class ItemThroughManager(models.Manager):
    def __init__(self, item_field_name, object_field_name):
        self._item_field_name = item_field_name
        self._object_field_name = object_field_name
        super(ItemThroughManager, self).__init__()

    @property
    def item_field(self):
        "Reference to the local foreign key reference to the `AbstractItem` subclass."
        return self.model._meta.get_field_by_name(self._item_field_name)[0]

    @property
    def object_field(self):
        "Reference to the local foreign key reference to the related object."
        return self.model._meta.get_field_by_name(self._object_field_name)[0]

    @property
    def index_model(self):
        "Reference to the item's `AbstractItemIndex` subclass."
        return self.item_field.rel.to.item_indexes.related.model

    def _construct_case_and_where(self, items, where_equals, node_join):
        """Constructs the SUM-CASE statement and corresponding WHERE clause
        for each item. WHERE clauses are set to `where_equals` and joined by
        the `node_join` operator, either AND or OR.
        """
        cases = []
        wheres = []

        if where_equals is True:
            predicate = ' {column_alias} > 0 '
        else:
            predicate = ' {column_alias} = {where_equals} '

        for pk in items:
            column_alias = COLUMN_PREFIX + str(pk)

            cases.append(SUMMED_CASE_TEMPLATE.format(**{
                'item_pk': pk,
                'index_table': qn(self.index_model._meta.db_table),
                'item_column': qn('item_id'),
                'parent_column': qn('parent_id'),
                'column_alias': qn(column_alias),
            }))

            wheres.append(predicate.format(column_alias=qn(column_alias),
                where_equals=where_equals))

        return ', '.join(cases), node_join.join(wheres)

    def _get_query(self, case_statements, where_conditions, evaluate, negate=False):
        """Constructs an SQL query based on the items, case_statments,
        and where conditions provided and runs the query on the associated
        model.
        """
        query = PIVOT_QUERY_TEMPLATE.format(**{
            'item_column': qn(self.item_field.column),
            'object_column': qn(self.object_field.column),
            'case_statements': case_statements,
            'through_table': qn(self.model._meta.db_table),
            'index_table': qn(self.index_model._meta.db_table),
            'where_condition': where_conditions,
            'item_id': qn('item_id'),
            'negate': negate and 'NOT' or '',
        })
        # Clean up the whitespace
        query = re.sub('\s+', ' ', query)
        if evaluate:
            return [x[0] for x in RawQuery(query, using=self.db)]
        return '({0})'.format(query)

    def _prepare_items(self, items):
        if type(items[0]) is not int:
            items = [term.pk for term in items]
        return items

    def _any(self, items, evaluate, negate):
        "Pivot query for disjunction."
        items = self._prepare_items(items)
        cases, wheres = self._construct_case_and_where(items, True, 'OR')
        return self._get_query(cases, wheres, evaluate, negate)

    def _all(self, items, evaluate, negate):
        "Pivot query for conjunction."
        items = self._prepare_items(items)
        cases, wheres = self._construct_case_and_where(items, True, 'AND')
        return self._get_query(cases, wheres, evaluate, negate)

    def _only(self, items, evaluate, negate):
        "Pivot query for exclusive conjunction."
        items = self._prepare_items(items)
        cases, wheres = self._construct_case_and_where(items, 1, 'AND')
        return self._get_query(cases, wheres, evaluate)

    def requires_any(self, items, evaluate=False):
        return self._any(items, evaluate, negate=False)

    def excludes_any(self, items, evaluate=False):
        return self._any(items, evaluate, negate=True)

    def requires_all(self, items, evaluate=False):
        return self._all(items, evaluate, negate=False)

    def excludes_all(self, items, evaluate=False):
        return self._all(items, evaluate, negate=True)

    def only(self, items, evaluate=False):
        return self._only(items, evaluate, negate=False)


class ItemIndexManager(models.Manager):
    def _index_ancestors(self, item, parent, depth=0):
        self.get_or_create(item=item, parent=parent, depth=depth)

        # recurse relative to next ancestor if exists
        if parent and parent.parent:
            self._index_ancestors(item, parent.parent, depth+1)

    @transaction.commit_on_success
    def index(self):
        indexed_model = self.model._meta.get_field_by_name('item')[0].rel.to
        items = indexed_model._default_manager.db_manager(self.db).all()

        for item in iter(items):
            self._index_ancestors(item, item.parent)

        # reset terminal flag
        items.update(terminal=False)
        indexes = self.get_query_set()

        # terminals are defined as having no children, thus the index will not
        # contain rows having ``item`` also be a ``parent``
        parents = indexes.exclude(parent=None).values_list('parent', flat=True)
        terminals = indexes.exclude(item__in=parents).distinct()\
            .values_list('item', flat=True)

        # update items that are terminals
        items.filter(pk__in=terminals).update(terminal=True)



from setuptools import setup, find_packages

kwargs = {
    # Packages
    'packages': find_packages(exclude=[
        'tests',
        '*.tests',
        '*.tests.*',
        'tests.*'
    ]),
    'include_package_data': True,

    # Dependencies
    'install_requires': [
        'serrano>=2.1.0,<2.4',
    ],

    # Test dependencies
    'tests_require': [
        'coverage',
    ],

    'test_suite': 'test_suite',

    # Metadata
    'name': 'harvest-vocab',
    'version': __import__('vocab').get_version(),
    'author': 'Byron Ruth',
    'author_email': 'b@devel.io',
    'description': 'A Harvest Stack app for modeling hierarchical data',
    'license': 'BSD',
    'keywords': 'django harvest avocado cilantro vocabulary hierarchical',
    'url': 'https://github.com/cbmi/harvest-vocab/',
    'classifiers': [
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Framework :: Django',
        'Topic :: Internet :: WWW/HTTP',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
    ],
}

setup(**kwargs)



from django.db.models import Q
from django.db import models, router
from vocab.managers import ItemManager, ItemIndexManager

class AbstractItem(models.Model):
    """Abstract class for a hierarchical item to be used with the vocab
    facilities.

    The foreign key or many-to-many `parent` field must be defined in concrete
    subclasses.

    The supplied `terminal` field is for operational purposes but can used as
    a shorthand way for finding all terminals/leaves in the tree. This value
    is set by the `ItemIndexManager` after an index is rebuilt.

    `description_field` is the name or description char/text field
    representative the item. This is used for ordering and textual
    representation of ID-based queries.

    `search_fields` are a list of fields that may be used for
    case-insensitive lookups (icontains). This is required for using the API
    resource classes.
    """
    terminal = models.NullBooleanField()

    objects = ItemManager()
    description_field = 'description'
    search_fields = ('description',)

    class Meta(object):
        abstract = True

    def __unicode__(self):
        return unicode(getattr(self, self.description_field))

    def ancestors(self, include_self=False):
        """Returns a `QuerySet` containing all ancestors of this item.

        If `include_self` is true, this item will be included in the query
        set.
        """
        db = router.db_for_read(self.__class__, instance=self)
        subquery = self.item_indexes.db_manager(db).exclude(parent=None).values_list('parent__pk', flat=True)
        if include_self:
            return self.__class__.objects.db_manager(db).filter(Q(pk__in=subquery) | Q(pk=self.pk))
        return self.__class__.objects.db_manager(db).filter(pk__in=subquery)

    def descendants(self, include_self=False):
        """Returns a `QuerySet` containing all descendants of this item.

        If `include_self` is true, this item will be included in the query
        set.
        """
        db = router.db_for_read(self.__class__, instance=self)
        subquery = self.parent_indexes.db_manager(db).values_list('item__pk', flat=True)
        if include_self:
            return self.__class__.objects.db_manager(db).filter(Q(pk__in=subquery) | Q(pk=self.pk))
        return self.__class__.objects.db_manager(db).filter(pk__in=subquery)


class AbstractItemIndex(models.Model):
    """Abstract class for an item index.

    Foreign keys named `item` and `parent` must be defined referencing an
    `AbstractItem` subclass.
    """
    depth = models.IntegerField(default=0)

    objects = ItemIndexManager()

    def __unicode__(self):
        return u'{0} via {1}'.format(self.item, self.parent)

    class Meta(object):
        abstract = True
        ordering = ('depth',)
